def size_conv_layer(s : int, kernel_size : int, padding : int, stride : int) -> int:
        '''
            Calculate the size of the image after convolution layer

            Parameters:  
            -----------  
            - s : int  
                integer represeting the size of one axis of the image  
            - kernel_size : int  
                integer represeting the size of the convolutional kernel  
            - padding : int  
                integer represeting the padding size  
            - stride : int  
                integer representing the stride size  
  
            Returns:  
            --------  
            - size : int  
                size after conv2D and Maxpool  
        '''

        size = int(((s - kernel_size + 2 * padding)/stride) + 1)
        return size


import sys
sys.path += ['.', './circuits/']

import pennylane as qml
import numpy as np

from hqm.circuits.circuit import QuantumCircuit

class Printer:
    '''
        This class contains static methods to print quantum circuit information.
    '''

    @staticmethod
    def draw_circuit(circuit : QuantumCircuit) -> str:
        '''
            Draw circuit structure.  

            Parameters:  
            -----------  
            - circuit : hqm.circuits.circuit.QuantumCircuit  
                hqm circuit to be drawn  
            
            Return:  
            -------  
            - str_circ : str  
                string containing circuit structure  

        '''
        
        weights  = np.random.random(size=circuit.weight_shape['weights'])
        inputs   = np.random.random(size=(circuit.n_qubits))
        str_circ = qml.draw(circuit.circuit, expansion_strategy="device")(inputs, weights)
        print(str_circ)
        return str_circ



import sys
sys.path += ['.', './hqm/']

from hqm.encoding.autoencoders import QuanvolutionAutoencoder, HybridAutoencoder
from hqm.circuits.angleencoding import BasicEntangledCircuit
from hqm.layers.basiclayer import BasicLayer

import torch

def test_quanvolution_autoencoder_output():

    qcircuits = []
    qcircuits += [BasicEntangledCircuit(n_qubits=9, n_layers=1)]
    qcircuits += [BasicEntangledCircuit(n_qubits=9, n_layers=1)]
    qcircuits += [BasicEntangledCircuit(n_qubits=9, n_layers=1)]

    autoencoder = QuanvolutionAutoencoder(qcircuits=qcircuits, 
                                            in_shape=(16,16,3), 
                                            filters=[3,6,9], 
                                            kernelsizes=[3,3,3], 
                                            strides=[1,1,1])
    
    x = torch.rand(size=(1,3,16,16))

    o = autoencoder(x)

    assert x.shape == o.shape

def test_hybrid_autoencoder_output():

    qcircuit  = BasicEntangledCircuit(n_qubits=9, n_layers=1)
    qlayer    = BasicLayer(qcircuit=qcircuit, aiframework='torch')

    autoencoder = HybridAutoencoder(qlayer=qlayer, 
                                    in_shape=(16,16,3), 
                                    filters=[3,6,9], 
                                    kernelsizes=[3,3,3], 
                                    strides=[1,1,1])
    
    x = torch.rand(size=(1,3,16,16))
    o = autoencoder(x)

    assert x.shape == o.shape
  







import sys
sys.path += ['.', './hqm/']

from hqm.circuits.angleencoding import StronglyEntangledCircuit
from hqm.regression.hmlp import BasicHybridMLPRegressor
from hqm.layers.basiclayer import BasicLayer

import matplotlib.pyplot as plt
import pennylane as qml
import torch

def test_qtorch_regressor_output():
    N_QUBITS      = 12
    N_LAYERS      = 4
    IN_DIM        = 20
    OU_DIM        = 20
    BATCH_SIZE    = 16

    dev = qml.device("default.qubit", wires=N_QUBITS)
    qcircuit = StronglyEntangledCircuit(n_qubits=N_QUBITS, n_layers=N_LAYERS, dev=dev)
    model = BasicHybridMLPRegressor(BasicLayer(qcircuit, aiframework='torch'), in_dim=IN_DIM, ou_dim=OU_DIM)
    
    x = torch.rand((BATCH_SIZE,   IN_DIM))
    o = model(x)

    assert o.shape == (BATCH_SIZE, OU_DIM)


import sys
sys.path += ['.', './layers/']

from hqm.layers.basiclayer import BasicLayer
import torch


class BasicHybridMLPRegressor(torch.nn.Module):
    '''
        This class implements a basic hybrid multilayer perceptron for regression purposes.
        BasicHybridMLPRegressor is composed of quantum layers stacked between two fully connected layers.
        The size of fully connected layers is set by means of in_dim and ou_dim.
    '''

    def __init__(self, qlayer : BasicLayer, in_dim : int, ou_dim : int) -> None:
        '''
            BasicHybridMLPRegressor constructor.  

            Parameters:  
            -----------  
            - qlayer : hqm.layers.basiclayer.BasicLayer
                hqm quantum layer to be stacked between two fully connected layers  
            - in_dim : int  
                integer representing the input size for the first fully connected layer  
            - ou_dim : int  
                integer representing the output size of the hybrid model  
            
            Returns:  
            --------  
            Nothing, a BasicHybridMLPRegressor object will be created.    
        '''

        super().__init__()

        if in_dim < 1: raise Exception(f"The parameter in_dim must be greater than 1, found {in_dim}")
        if ou_dim < 1: raise Exception(f"The parameter ou_dim must be greater than 1, found {ou_dim}")

        n_qubits  = qlayer.n_qubits
        self.fc_1 = torch.nn.Linear(in_dim, n_qubits)
        self.qc_1 = qlayer.qlayer
        self.fc_2 = torch.nn.Linear(n_qubits, ou_dim)
        self.tanh = torch.nn.Tanh()

    def forward(self, x : torch.Tensor) -> torch.Tensor:
        ''' 
            Torch forward method  

            Parameters:  
            -----------  
            - x : torch.Tensor  
                input for the torch model  

            Returns:  
            --------  
            - out : torch.Tensor  
                output from the torch model  
        '''

        x = self.fc_1(x)
        x = self.tanh(x)
        x = self.qc_1(x)
        x = self.tanh(x)
        x = self.fc_2(x)
        out = self.tanh(x)
        return out
    
class MultiHybridMLPRegressor(torch.nn.Module):
    '''
        This class implements a hybrid multilayer perceptron with multiple quantum circuits for regression purposes.
        MultiHybridMLPRegressor is composed of several quantum layers stacked between two fully connected layers.
        The size of fully connected layers is set by means of in_dim and ou_dim.
    '''

    def __init__(self, qlayers : list, in_dim : int, ou_dim : int) -> None:
        '''
            MultiHybridMLPRegressor constructor.  

            Parameters:  
            -----------  
            - qlayer : list  
                list of hqm quantum layers to be stacked between two fully connected layers  
            - in_dim : int  
                integer representing the input size for the first fully connected layer  
            - ou_dim : int  
                integer representing the output size of the hybrid model  
            
            Returns:  
            --------  
            Nothing, a MultiHybridMLPRegressor object will be created.    
        '''

        super().__init__()

        if in_dim < 1: raise Exception(f"The parameter in_dim must be greater than 1, found {in_dim}")
        if ou_dim < 1: raise Exception(f"The parameter ou_dim must be greater than 1, found {ou_dim}")
        if len(qlayers) < 1: raise Exception(f"Size of qlayers must be greater than 1, found {len(qlayers)}")

        n_qubits_0 = qlayers[0].n_qubits
        n_qubits_1 = qlayers[-1].n_qubits
        self.fc_1  = torch.nn.Linear(in_dim, n_qubits_0)
        self.qcs   = [circ.qlayer for circ in qlayers]
        self.fc_2  = torch.nn.Linear(n_qubits_1, ou_dim)
        self.tanh  = torch.nn.Tanh()

    def forward(self, x : torch.Tensor) -> torch.Tensor:
        '''
            Torch forward method  

            Parameters:  
            -----------  
            - x : torch.Tensor  
                input for the torch model  

            Returns:  
            --------  
            - out : torch.Tensor  
                output from the torch model  
        '''

        x = self.fc_1(x)
        x = self.tanh(x)
        for qc in self.qcs: 
            x = qc(x)
            x = self.tanh(x)
        x = self.fc_2(x)
        out = self.tanh(x)
        return out

class MultiHybridMultiMLPRegressor(torch.nn.Module):
    '''
        This class implements a hybrid multilayer perceptron with multiple quantum circuits for regression purposes.
        MultiHybridMultiMLPRegressor is composed of several quantum layers stacked between alternating fully connected layers.
        The size of fully connected layers is set by means of in_dim and ou_dim.
    '''

    def __init__(self, qlayers : list, in_dims : list, ou_dim : list) -> None:
        '''
            MultiHybridMultiMLPRegressor constructor.  

            Parameters:  
            -----------  
            - qlayers : list  
                list of hqm quantum qlayerss to be stacked between two fully connected layers  
            - in_dims: list  
                list of integers representing the input size for the i-th fully connected layer (first value should correspond to size of input data)  
            - ou_dim : list  
                list of integers representing the output size for the i-th fully connected layer (last value should correspond to desired output size)  
            
            Returns:  
            --------  
            Nothing, a MultiHybridMLPRegressor object will be created.    
        '''
        
        super().__init__()

        if len(in_dims) < 1: raise Exception(f"Size in_dims must be greater than 1, found {len(in_dims)}")
        if ou_dim < 1: raise Exception(f"The parameter ou_dim must be greater than 1, found {ou_dim}")
        if len(qlayers) < 1: raise Exception(f"Size of qlayers must be greater than 1, found {len(qlayers)}")
        if len(qlayers) != len(in_dims): raise Exception(f"qlayers and in_dims must have the same lenght, found {len(qlayers)} and {len(in_dims)}")
        for i, dim in enumerate(in_dims): 
            if dim < 1: raise Exception(f"Element {i} of in_dims must be greater than 1, found {dim}")
            else: pass
        
        self.fcs  = [torch.nn.Linear(dim, circ.n_qubits) for (dim, circ) in zip(in_dims, qlayers)]
        self.fco  = torch.nn.Linear(qlayers[-1].n_qubits, ou_dim)
        self.qcs  = [circ.qlayer for circ in qlayers]
        self.tanh = torch.nn.Tanh()

    def forward(self, x : torch.Tensor) -> torch.Tensor:
        '''
            Torch forward method  

            Parameters:  
            -----------  
            - x : torch.Tensor  
                input for the torch model  

            Returns:  
            --------  
            - x : torch.Tensor  
                output from the torch model  
        '''
        
        for fc, qc in zip(self.fcs, self.qcs):
            x = fc(x)
            x = self.tanh(x)
            x = qc(x)
            x = self.tanh(x)
        
        x   = self.fco(x) 
        out = self.tanh(x)
        return out


import sys
sys.path += ['.', './hqm/']

from hqm.circuits.angleencoding import BasicEntangledCircuit
from hqm.layers.quanvolution import Quanvolution2D
from hqm.utils.printer import Printer

import pennylane as qml
import numpy as np
import argparse
import torch


def test_quanvolution_padding_same_output():
    N_QUBITS      = 4
    N_LAYERS      = 1
    FITLERS       = 4
    KERNELSIZE    = 2
    STRIDE        = 1
    IMG_SIZE      = 16
    BATCH_SIZE    = 1
    CHANNELS      = 1
    PADDING       = 'same'

    #=======================================================================
    # InizializeQuanvolution2D
    #=======================================================================
    print('Initializing hybrid model', '\n')
    dev = qml.device("lightning.qubit", wires=N_QUBITS)

    qcircuit = BasicEntangledCircuit(n_qubits=N_QUBITS, n_layers=N_LAYERS, dev=dev)
    quanv    = Quanvolution2D(qcircuit=qcircuit, filters=FITLERS, kernelsize=KERNELSIZE, stride=STRIDE, padding=PADDING)

    #=======================================================================
    # Applying Quanvolution2D - padding same
    #=======================================================================

    x = torch.rand(size=(BATCH_SIZE,CHANNELS,IMG_SIZE,IMG_SIZE))
    o =  quanv(x)

    assert o.shape == (BATCH_SIZE, FITLERS, IMG_SIZE, IMG_SIZE)

def test_quanvolution_padding_valid_output():
    N_QUBITS      = 9
    N_LAYERS      = 1
    FITLERS       = 9
    KERNELSIZE    = 3
    STRIDE        = 1
    IMG_SIZE      = 16
    BATCH_SIZE    = 2
    CHANNELS      = 3
    PADDING       = 'valid'

    #=======================================================================
    # InizializeQuanvolution2D
    #=======================================================================
    print('Initializing hybrid model', '\n')
    dev = qml.device("lightning.qubit", wires=N_QUBITS)

    qcircuit = BasicEntangledCircuit(n_qubits=N_QUBITS, n_layers=N_LAYERS, dev=dev)
    quanv    = Quanvolution2D(qcircuit=qcircuit, filters=FITLERS, kernelsize=KERNELSIZE, stride=STRIDE, padding=PADDING)

    #=======================================================================
    # Applying Quanvolution2D - padding same
    #=======================================================================

    x = torch.rand(size=(BATCH_SIZE,CHANNELS,IMG_SIZE,IMG_SIZE))
    o =  quanv(x)

    assert o.shape == (BATCH_SIZE, FITLERS, int((IMG_SIZE - KERNELSIZE)/STRIDE + 1), int((IMG_SIZE - KERNELSIZE)/STRIDE + 1))
    


import pennylane as qml
import typing

class AIInterface:
    '''
        This class implements the inteface between quantum circuits implemented in PennyLane and the two most popular DL libraries,
        Pytorch and Tensorflow. 
    '''

    @staticmethod
    def network_layer(circuit : qml.qnode, weight_shape : dict, n_qubits : int, aiframework : str) -> typing.Union[qml.qnn.TorchLayer, qml.qnn.KerasLayer]:
        '''
            Static methods that embedd quantum layer into a torch or a keras layer.  

            Parameters:  
            -----------  
            - circuit : qml.qnode  
                pennylane circuit to be embedded   
            - weight_shape : dict  
                shape of the trainalbe weights, it is derived from hqm.circuits.circuit.QuantumCircuit  
            - n_qubits : int  
                integer representing the number of qubits used for the circuit  
            - aiframeworkks : str  
                string representing which in wich ai framework the circuit will be embedded, can be 'torch' or 'keras'  

            Returns:   
            --------  
            - qlayer : qml.qnn.TorchLayer or qml.qnn.KerasLayer  
        '''

        if   aiframework not in ['torch', 'keras']: raise Exception(f"Accepted values for framerwork are 'torch' or 'keras', found {aiframework}")
        if   aiframework == "torch": qlayer = qml.qnn.TorchLayer(circuit, weight_shape)
        elif aiframework == "keras": qlayer = qml.qnn.KerasLayer(circuit, weight_shape, output_dim=n_qubits)
        else: raise Exception(f"Framerwork can be only torch or keras, found {aiframework}!")
        
        return qlayer
